

#include <Astri_Redis_DAQ_Observer.h>
#include <unistd.h>
#include <chrono>
#include <ctime>
#include <iomanip>

using namespace inaf::oasbo::DAQ_observers;

std::string AstriRedisDAQObserver::lastopenfile_key = "last_open_file";
std::string AstriRedisDAQObserver::lastclosedfile_key = "last_closed_file";
std::string AstriRedisDAQObserver::mode_key = "daq_mode";
std::string AstriRedisDAQObserver::receiverSource_key = "receiver_source";
std::string AstriRedisDAQObserver::receiverConn_key = "receiver_connection";
std::string AstriRedisDAQObserver::provider_destination_key =
		"provider_destination";

AstriRedisDAQObserver::AstriRedisDAQObserver(inaf::oasbo::DAQ::BaseDAQ &daq,
		std::string ip, int port) :
		inaf::oasbo::DAQ_observers::BaseDAQ_Observer(daq) {

	astri_daq = static_cast<inaf::oasbo::DAQ::AstriDAQ*>(&daq);
	this->updateInterval = 2;
	this->ip = ip;
	this->port = port;

	keys.insert(lastopenfile_key);
	keys.insert(lastclosedfile_key);
	keys.insert(mode_key);
	keys.insert(receiverSource_key);
	keys.insert(receiverConn_key);
	keys.insert(provider_destination_key);
}

AstriRedisDAQObserver::~AstriRedisDAQObserver() {
	BaseDAQ_Observer::~BaseDAQ_Observer();
	redisFree(context);
}

void AstriRedisDAQObserver::updateDAQStats() {
	if (!isConnected())
		return;
	if (this->writeMutex.try_lock()) {
		std::string state(
				this->astri_daq->getStateStr(
						this->astri_daq->getCurrentState()));
		redisReply *r = (redisReply*) redisCommand(context, "SET %s %s",
				mode_key.c_str(), state.c_str());
		this->writeMutex.unlock();
		if (r == NULL) {
			return;
		}
		// Check the type of the Redis reply
		if (r->type == REDIS_REPLY_ERROR) {
			// Error executing Redis command
			time_t now = time(nullptr);
			std::cerr << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[Redis Observer]\t" << r->str << std::endl;
			freeReplyObject(r);
			return;
		}
		freeReplyObject(r);
	}
}
void AstriRedisDAQObserver::updatePacketStats() {
	if (!isConnected())
		return;
	inaf::oasbo::PacketMonitors::BasePacketMonitor &monitor =
			*(this->dataAcquisition->getMonitorPtr());
	for (const auto& [key, value] : monitor.getStatsMap()) {
		keys.insert(key);
		if (this->writeMutex.try_lock()) {
			redisReply *r = (redisReply*) redisCommand(context, "SET %s %s",
					key.c_str(), value.c_str());
			this->writeMutex.unlock();
			if (r == NULL) {
				return;
			}
			// Check the type of the Redis reply
			if (r->type == REDIS_REPLY_ERROR) {
				// Error executing Redis command
				time_t now = time(nullptr);
				std::cerr << "["
						<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
						<< "]\t[Redis Observer]\t" << r->str << std::endl;
				freeReplyObject(r);
				return;
			}
			freeReplyObject(r);
		}

	}
}

void AstriRedisDAQObserver::updateArchiverStats() {
	if (!isConnected())
		return;
	std::string dest = this->dataAcquisition->getArchiverPtr()->getDest();

	std::string cmd;
	if (this->dataAcquisition->getArchiverPtr()->is_open())
		cmd = lastopenfile_key;
	else {
		cmd = lastclosedfile_key;
	}
	if (this->writeMutex.try_lock()) {
		redisReply *r = (redisReply*) redisCommand(context, "SET %s %s",
				cmd.c_str(), dest.c_str());
		this->writeMutex.unlock();
		if (r == NULL) {
			return;
		}
		// Check the type of the Redis reply
		if (r->type == REDIS_REPLY_ERROR) {
			// Error executing Redis command
			time_t now = time(nullptr);
			std::cerr << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[Redis Observer]\t" << r->str << std::endl;
			freeReplyObject(r);
			return;
		}
		freeReplyObject(r);
	}

}

void AstriRedisDAQObserver::updateProviderStats() {
	if (!isConnected())
		return;
	std::string dest = this->dataAcquisition->getProviderPtr()->getDest();
	if (this->writeMutex.try_lock()) {
		redisReply *r = (redisReply*) redisCommand(context, "SET %s %s",
				provider_destination_key.c_str(), dest.c_str());
		this->writeMutex.unlock();
		if (r == NULL) {
			return;
		}
		// Check the type of the Redis reply
		if (r->type == REDIS_REPLY_ERROR) {
			// Error executing Redis command
			time_t now = time(nullptr);
			std::cerr << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[Redis Observer]\t" << r->str << std::endl;
			freeReplyObject(r);
			return;
		}
		freeReplyObject(r);
	}
}

void AstriRedisDAQObserver::updateReceiverStats() {
	if (!isConnected())
		return;
	std::string dest = this->dataAcquisition->getReceiverPtr()->getHost();
	std::string connected =
			this->dataAcquisition->getReceiverPtr()->isConnectedToClient() ?
					"Connected" : "Not connected";
	if (this->writeMutex.try_lock()) {
		redisReply *r = (redisReply*) redisCommand(context, "SET %s %s",
				receiverSource_key.c_str(), dest.c_str());

		if (r == NULL) {
			this->writeMutex.unlock();
			return;
		}
		// Check the type of the Redis reply
		if (r->type == REDIS_REPLY_ERROR) {
			this->writeMutex.unlock();
			// Error executing Redis command
			time_t now = time(nullptr);
			std::cerr << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[Redis Observer]\t" << r->str << std::endl;
			freeReplyObject(r);
			return;
		}
		freeReplyObject(r);
		r = (redisReply*) redisCommand(context, "SET %s %s",
				receiverConn_key.c_str(), connected.c_str());

		if (r == NULL) {
			this->writeMutex.unlock();
			return;
		}
		// Check the type of the Redis reply
		if (r->type == REDIS_REPLY_ERROR) {
			this->writeMutex.unlock();
			// Error executing Redis command
			time_t now = time(nullptr);
			std::cerr << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[Redis Observer]\t" << r->str << std::endl;
			freeReplyObject(r);
			return;
		}
		freeReplyObject(r);
		this->writeMutex.unlock();

	}
}

void AstriRedisDAQObserver::updateAll() {
	updateArchiverStats();
	updatePacketStats();
	updateProviderStats();
	updateReceiverStats();
	updateDAQStats();
}

void AstriRedisDAQObserver::updateThread() {
	this->stopFlag = false;
	while (!this->stopFlag) {
		updateAll();
		std::unique_lock<std::mutex> lk(this->writeMutex);
		this->sleep_cv.wait_for(lk, std::chrono::seconds(this->updateInterval),
				[&] {
					return this->stopFlag;
				}); // wait for "updateInterval" seconds or until notified
	}
}

void AstriRedisDAQObserver::start() {
	if (this->observerThread.joinable()) {
		return;
	}
	this->observerThread = std::thread(&AstriRedisDAQObserver::updateThread,
			this);
	if (this->checkConn.joinable()) {
		return;
	}
	this->checkConnThreadFlag = true;
	this->context = redisConnect(ip.c_str(), port);
	if (isConnected()) {
		time_t now = time(nullptr);
		std::cout << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[Redis Observer]" << "\tConnected" << std::endl;
	}
	this->checkConn = std::thread(&AstriRedisDAQObserver::connectLoop, this);
}

void AstriRedisDAQObserver::stop() {
	if (this->observerThread.joinable()) {
		this->stopFlag = true;
		this->sleep_cv.notify_all();
		this->observerThread.join();
	}
	for (auto key : keys) {
		if (!isConnected()) {
			continue;
		}
		redisReply *r = (redisReply*) redisCommand(context, "SET %s 0",
				key.c_str());
		if (r == NULL) {
			return;
		}
		// Check the type of the Redis reply
		if (r->type == REDIS_REPLY_ERROR) {
			// Error executing Redis command
			time_t now = time(nullptr);
			std::cerr << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[Redis Observer]\t" << r->str << std::endl;
			freeReplyObject(r);
			return;
		}
		freeReplyObject(r);
	}
	this->checkConnThreadFlag = false;
	if (this->checkConn.joinable()) {
		this->checkConn.join();
	}
}

bool AstriRedisDAQObserver::isConnected() {
	return (context != nullptr && !context->err);
}

int AstriRedisDAQObserver::connectLoop() {
	while (checkConnThreadFlag) {
		if (!isConnected()) {
			time_t now = time(nullptr);
			std::cerr << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[Redis Observer]" << "\tNot connected" << std::endl;
			this->context = redisConnect(ip.c_str(), port);
			if (isConnected()) {
				time_t now = time(nullptr);
				std::cout << "["
						<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
						<< "]\t[Redis Observer]" << "\tConnected" << std::endl;
			}
		}
		std::this_thread::sleep_for(std::chrono::seconds(1));
	}
	return 1;
}
